### Author: Dag Wieers <dag$wieers,com>

class dool_plugin(dool):
    """
    Number of active dbus sessions.
    """
    def __init__(self):
        self.name = 'dbus'
        self.nick = ('sys', 'ses')
        self.vars = ('system', 'session')
        self.type = 'd'
        self.width = 3
        self.scale = 100

    def check(self):
#       dstat.info(1, 'The dbus module is an EXPERIMENTAL module.')
        try:
            global dbus
            import dbus
            try:
                self.sysbus = dbus.Bus(dbus.Bus.TYPE_SYSTEM).get_service('org.freedesktop.DBus').get_object('/org/freedesktop/DBus', 'org.freedesktop.DBus')
                try:
                    self.sesbus = dbus.Bus(dbus.Bus.TYPE_SESSION).get_service('org.freedesktop.DBus').get_object('/org/freedesktop/DBus', 'org.freedesktop.DBus')
                except:
                    self.sesbus = None
            except:
                raise Exception('Unable to connect to dbus message bus')
        except:
            raise Exception('Needs python-dbus module')

    def extract(self):
        self.val['system'] = len(self.sysbus.ListServices()) - 1
        try:
            self.val['session'] = len(self.sesbus.ListServices()) - 1
        except:
            self.val['session'] = -1
#       print(dir(b)); print(dir(s)); print(dir(d)); print(d.ListServices())
#       print(dir(d))
#       print(d.ListServices())

# vim:ts=4:sw=4:et



### Author: Dag Wieers <dag@wieers.com>

class dool_plugin(dool):
    """
    Total Number of processes on this system.
    """
    def __init__(self):
        self.name = 'procs'
        self.vars = ('total',)
        self.type = 'd'
        self.width = 4
        self.scale = 10

    def extract(self):
        self.val['total'] = len([pid for pid in proc_pidlist()])



### Author: Scott Baker - <https://www.perturb.org/>

class dool_plugin(dool):
	"""
	Plugin to gather data about a specific PID
	"""

	def __init__(self):
		self.vars  = ('cpu  read  writ  mem',)
		self.type  = 's'
		self.width = 20
		self.scale = 0
		self.prev  = { 'read': -1, 'write': -1, 'cpu': 0, 'mem': 0}

		# Get the pid from the CLI options
		try:
			pid  = int(op.plugin_params['pid-detail'])
		except:
			msg = 'pid-detail: %s is not a number' % (op.plugin_params['pid-detail'])
			raise Exception(msg)

		path = '/proc/%s/stat' % pid
		if not os.path.exists(path):
			msg = 'pid-detail: %d is not an active pid' % (pid)
			raise Exception(msg)

		name = getnamebypid(pid, 'unknown')

		# Set the plugin title at the top of the column
		self.name = "pid info: %s" % name

		# Save the pid name and number for easier access later
		self.pid_name = name
		self.pid      = pid

	def extract(self):
		pid = self.pid

		# Make sure we have data for this pid
		path = '/proc/%s/stat' % pid
		if not os.path.exists(path):
			self.output = 'Pid #%s not found' % pid
			return

		(disk_rd, disk_wr) = self.get_pid_disk(pid)
		cpu_usage = self.get_pid_cpu(pid)
		mem_usage = self.get_pid_mem(pid)

		cpu_str   = cprint(cpu_usage, 'f', 3, 34)
		disk_rd   = cprint(disk_rd	, 'd', 5, 1024)
		disk_wr   = cprint(disk_wr	, 'd', 5, 1024)
		mem_usage = cprint(mem_usage, 'd', 4, 1024)

		self.output = "%s %s %s %s" % (cpu_str, disk_rd, disk_wr, mem_usage)

	# Sometimes /proc/12345/stat has a space in the proc name field
	# which messes up the split. This tries to be smart and work around that
	def get_proc_pid_stat(self, pid):
		path = '/proc/%s/stat' % pid
		if not os.path.exists(path):
			return []

		# Read the first line
		line = proc_readline(path)
		line = line.strip()

		# Find the proc name that's between the ( )
		x     = re.search("\((.*?)\)", line)
		paren = x.group(1)

		# Replace any spaces in the parens part with underscores
		cleaned  = paren.replace(" ", "_")
		new_line = line.replace(paren, cleaned, 1)

		# Return the split array
		parts = new_line.split(" ")

		return parts

	def get_pid_cpu(self, pid):
		path = '/proc/%s/stat' % pid
		if not os.path.exists(path):
			self.output = 'Pid #%s not found' % pid

		l = self.get_proc_pid_stat(pid)
		if len(l) < 35: return 0

		user_time   = int(l[13])
		kernel_time = int(l[14])

		current_cpu = user_time + kernel_time
		prev_cpu    = self.prev['cpu']
		usage       = (current_cpu - prev_cpu) * 1.0 / elapsed / cpunr

		if prev_cpu:
			ret = usage
		else:
			ret = 0

		# If we're at a real update (not interim) we save the data
		if step == op.delay:
			self.prev['cpu'] = current_cpu

		return ret

	def get_pid_disk(self, pid):
		path  = '/proc/%s/io' % pid
		count = 0
		for p in proc_splitlines(path):
			key = p[0]
			val = p[1]

			if key == "read_bytes:":
				cur_read = int(val)
			elif key == "write_bytes:":
				cur_write = int(val)

			count += 1

		# This is usually a permission error reading that part of /proc/
		if count == 0:
			return (0,0)

		if (op.bits):
			factor = 8
		else:
			factor = 1

		# We have previous data so we calculate the difference
		# between the two
		if self.prev['read'] > -1:
			rread	= (cur_read - self.prev['read'])   * factor / elapsed
			rwrite	= (cur_write - self.prev['write']) * factor / elapsed
		# First round we just gather data
		else:
			rread  = 0
			rwrite = 0

		# If we're at a real update (not interim) we save the data
		if step == op.delay:
			self.prev['read']  = cur_read
			self.prev['write'] = cur_write

		ret = (rread, rwrite)

		return ret

	def get_pid_mem(self, pid):
		l = proc_splitline('/proc/%s/statm' % pid)

		if len(l) < 2: return 0

		prev_usage = self.prev['mem']
		cur_usage  = (int(l[1]) * pagesize)

		# If we have previous numbers we return the average
		if prev_usage:
			usage = self.prev['mem'] / elapsed
		# First run just gets the current number
		else:
			usage = cur_usage

		self.prev['mem'] += cur_usage

		if step == op.delay:
			self.prev['mem'] = cur_usage

		# Uncomment this to always return the absolute (not average)
		#usage = cur_usage

		return usage

	def showcsv(self):
		xstr = "%s %s %s %s" % (1,2,3,4)

		return xstr

# vim: tabstop=4 shiftwidth=4 noexpandtab autoindent softtabstop=4



### Author: Dag Wieers <dag@wieers.com>

global socket
import socket

global struct
import struct

### FIXME: Implement millisecond granularity as well
### FIXME: Interrupts socket if data is overdue (more than 250ms ?)

class dool_plugin(dool):
    """
    Time from an NTP server.

    BEWARE: this dstat plugin typically takes a lot longer to run than
    system plugins and for that reason it is important to use an NTP server
    located nearby as well as make sure that it does not impact your other
    counters too much.
    """

    def __init__(self):
        self.name      = 'ntp'
        self.nick      = ('date/time',)
        self.vars      = ('time',)
        self.timefmt   = os.getenv('DOOL_TIMEFMT') or '%d-%m %H:%M:%S'
        self.ntpserver = os.getenv('DOOL_NTPSERVER') or '0.fedora.pool.ntp.org'
        self.type      = 's'
        self.width     = len(time.strftime(self.timefmt, time.localtime()))
        self.scale     = 0
        self.epoch     = 2208988800
#        socket.setdefaulttimeout(0.25)
        self.socket = socket.socket( socket.AF_INET, socket.SOCK_DGRAM )
        self.socket.settimeout(0.25)

        if op.debug:
            print("NTP Plugin is using: '%s'" % self.ntpserver)

    def gettime(self):
        self.socket.sendto( b'\x1b' + 47 * b'\0', ( self.ntpserver, 123 ))
        data, address = self.socket.recvfrom(1024)
        return struct.unpack( '!12I', data )[10] - self.epoch

    def check(self):
        try:
            self.gettime()
        except socket.gaierror:
            raise Exception('Failed to connect to NTP server %s.' % self.ntpserver)
        except socket.error:
            raise Exception('Error connecting to NTP server %s.' % self.ntpserver)

    def extract(self):
        try:
            self.val['time'] = time.strftime(self.timefmt, time.localtime(self.gettime()))
        except:
            self.val['time'] = theme['error'] + '-'.rjust(self.width-1) + ' '

    def showcsv(self):
        return time.strftime(self.timefmt, time.localtime(self.gettime()))

# vim:ts=4:sw=4:et



### Author: Dag Wieers <dag$wieers,com>

class dool_plugin(dool):
    """
    Number of operations performed on a GPFS filesystem.
    """

    def __init__(self):
        self.name = 'gpfs file operations'
        self.nick = ('open', 'clos', 'read', 'writ', 'rdir', 'inod')
        self.vars = ('_oc_', '_cc_', '_rdc_', '_wc_', '_dir_', '_iu_')
        self.type = 'd'
        self.width = 5
        self.scale = 1000

    def check(self): 
        if os.access('/usr/lpp/mmfs/bin/mmpmon', os.X_OK):
            try:
                self.stdin, self.stdout, self.stderr = dpopen('/usr/lpp/mmfs/bin/mmpmon -p -s')
                self.stdin.write('reset\n')
                readpipe(self.stdout)
            except IOError:
                raise Exception('Cannot interface with gpfs mmpmon binary')
            return True
        raise Exception('Needs GPFS mmpmon binary')

    def extract(self):
        try:
            self.stdin.write('io_s\n')
#           readpipe(self.stderr)
            for line in readpipe(self.stdout):
                if not line: continue
                l = line.split()
                for name in self.vars:
                    self.set2[name] = int(l[l.index(name)+1])
            for name in self.vars:
                self.val[name] = (self.set2[name] - self.set1[name]) * 1.0 / elapsed
        except IOError as e:
            if op.debug > 1: print('%s: lost pipe to mmpmon, %s' % (self.filename, e))
            for name in self.vars: self.val[name] = -1
        except Exception as e:
            if op.debug > 1: print('%s: exception %s' % (self.filename, e))
            for name in self.vars: self.val[name] = -1

        if step == op.delay:
            self.set1.update(self.set2)

# vim:ts=4:sw=4:et



### Author: Dag Wieers <dag@wieers.com>

class dool_plugin(dool):
    def __init__(self):
        self.name = 'thermal'
        self.type = 'd'
        self.width = 3
        self.scale = 20

        if os.path.exists('/sys/devices/virtual/thermal/'):
            self.nick = []
            self.vars = []
            for zone in os.listdir('/sys/devices/virtual/thermal/'):
                zone_split=zone.split("thermal_zone")
                if len(zone_split) == 2:
                    self.vars.append(zone)
                    name="".join(["tz",zone_split[1]])
                    self.nick.append(name)

        elif os.path.exists('/sys/bus/acpi/devices/LNXTHERM:01/thermal_zone/'):
            self.vars = os.listdir('/sys/bus/acpi/devices/LNXTHERM:01/thermal_zone/')
            self.nick = []
            for name in self.vars:
                self.nick.append(name.lower())

        elif os.path.exists('/proc/acpi/ibm/thermal'):
            self.namelist = ['cpu', 'pci', 'hdd', 'cpu', 'ba0', 'unk', 'ba1', 'unk']
            self.nick = []
            for line in dopen('/proc/acpi/ibm/thermal'):
                l = line.split()
                for i, name in enumerate(self.namelist):
                    if int(l[i+1]) > 0:
                        self.nick.append(name)
            self.vars = self.nick

        elif os.path.exists('/proc/acpi/thermal_zone/'):
            self.vars = os.listdir('/proc/acpi/thermal_zone/')
#           self.nick = [name.lower() for name in self.vars]
            self.nick = []
            for name in self.vars:
                self.nick.append(name.lower())

        else:
            raise Exception('Needs kernel thermal, ACPI or IBM-ACPI support')

    def check(self):
        if not os.path.exists('/proc/acpi/ibm/thermal') and \
           not os.path.exists('/proc/acpi/thermal_zone/') and \
           not os.path.exists('/sys/devices/virtual/thermal/') and \
           not os.path.exists('/sys/bus/acpi/devices/LNXTHERM:00/thermal_zone/'):
            raise Exception('Needs kernel thermal, ACPI or IBM-ACPI support')

    def extract(self):
        if os.path.exists('/sys/devices/virtual/thermal/'):
            for zone in self.vars:
                for line in dopen('/sys/devices/virtual/thermal/'+zone+'/temp').readlines():
                    l = line.split()
                    self.val[zone] = int(l[0])
        elif os.path.exists('/sys/bus/acpi/devices/LNXTHERM:01/thermal_zone/'):
            for zone in self.vars:
                if os.path.isdir('/sys/bus/acpi/devices/LNXTHERM:01/thermal_zone/'+zone) == False:
                    for line in dopen('/sys/bus/acpi/devices/LNXTHERM:01/thermal_zone/'+zone).readlines():
                        l = line.split()
                        if l[0].isdigit() == True:
                            self.val[zone] = int(l[0])
                        else:
                            self.val[zone] = 0
        elif os.path.exists('/proc/acpi/ibm/thermal'):
            for line in dopen('/proc/acpi/ibm/thermal'):
                l = line.split()
                for i, name in enumerate(self.namelist):
                    if int(l[i+1]) > 0:
                        self.val[name] = int(l[i+1])
        elif os.path.exists('/proc/acpi/thermal_zone/'):
            for zone in self.vars:
                for line in dopen('/proc/acpi/thermal_zone/'+zone+'/temperature').readlines():
                    l = line.split()
                    self.val[zone] = int(l[1])

# vim:ts=4:sw=4:et



### Author: Dag Wieers <dag$wieers,com>

class dool_plugin(dool):
    """
    Show internal Dool step and delay counters
    """

    def __init__(self):
        self.name = 'dool'
        #self.nick = ('counter',)
        self.vars = ('stp','dly','cnt',)
        self.type = 'd'
        self.width = 3
        self.scale = 10

    def extract(self):
        self.val['stp']  = step
        self.val['dly'] = op.delay

        if op.count == -1:
            xcount = 0
        else:
            xcount = op.count

        self.val['cnt'] = xcount

# vim:ts=4:sw=4:et



### Author: Dag Wieers <dag$wieers,com>

class dool_plugin(dool):
    def __init__(self):
        self.name = 'load avg'
        self.nick = ('1m', '5m', '15m')
        self.vars = ('load1', 'load5', 'load15')
        self.type = 'f'
        self.width = 4
        self.scale = 0.5
        self.server = os.getenv('DOOL_SNMPSERVER') or '192.168.1.1'
        self.community = os.getenv('DOOL_SNMPCOMMUNITY') or 'public'

    def check(self):
        try:
            global cmdgen
            from pysnmp.entity.rfc3413.oneliner import cmdgen
        except:
            raise Exception('Needs pysnmp and pyasn1 modules')

    def extract(self):
        list(map(lambda x, y: self.val.update({x: float(y)}), self.vars, snmpwalk(self.server, self.community, (1,3,6,1,4,1,2021,10,1,3))))

# vim:ts=4:sw=4:et



### Author: Dag Wieers <dag$wieers,com>

class dool_plugin(dool):
    """
    Number of read and write transactions per device.

    Displays the number of read and write I/O transactions per device.
    """

    def __init__(self):
        self.nick = ('#read', '#writ' )
        self.type = 'd'
        self.width = 5
        self.scale = 1000
        self.open('/proc/diskstats')
        self.cols = 2

    def discover(self, *objlist):
        ret = []
        for l in self.splitlines():
            if len(l) < 13: continue
            if set(l[3:]) == {'0'}: continue
            name = l[2]
            ret.append(name)
        for item in objlist: ret.append(item)
        if not ret:
            raise Exception('No suitable block devices found to monitor')
        return ret

    def vars(self):
        ret = []
        if op.disklist:
            varlist = op.disklist
        elif not op.full:
            varlist = ('total',)
        else:
            varlist = []
            for name in self.discover:
                if DOOL_DISKFILTER.match(name): continue
                if name not in blockdevices(): continue
                varlist.append(name)
#           if len(varlist) > 2: varlist = varlist[0:2]
            varlist.sort()
        for name in varlist:
            if name in self.discover + ['total'] or name in op.diskset:
                ret.append(name)
        return ret

    def name(self):
        return ['dsk/'+sysfs_dev(name) for name in self.vars]

    def extract(self):
        for name in self.vars: self.set2[name] = (0, 0)
        for l in self.splitlines():
            if len(l) < 13: continue
            if l[3] == '0' and l[7] == '0': continue
            if set(l[3:]) == {'0'}: continue
            name = l[2]
            if not DOOL_DISKFILTER.match(name):
                self.set2['total'] = ( self.set2['total'][0] + int(l[3]), self.set2['total'][1] + int(l[7]) )
            if name in self.vars and name != 'total':
                self.set2[name] = ( self.set2[name][0] + int(l[3]), self.set2[name][1] + int(l[7]))
            for diskset in self.vars:
                if diskset in op.diskset:
                    for disk in op.diskset[diskset]:
                        if re.match('^'+disk+'$', name):
                            self.set2[diskset] = ( self.set2[diskset][0] + int(l[3]), self.set2[diskset][1] + int(l[7]) )

        for name in self.set2:
            self.val[name] = list(map(lambda x, y: (y - x) / elapsed, self.set1[name], self.set2[name]))

        if step == op.delay:
            self.set1.update(self.set2)



### Author: Scott Baker <scott@perturb.org>

class dool_plugin(dool):
    def __init__(self):
        self.name    = 'memory percent'
        self.nick    = ('used', 'free', 'cach', 'avai')
        self.vars    = ('MemUsed', 'MemFree', 'Cached', 'MemAvailable')
        self.type    = 's'
        self.scale   = 0
        self.num_fmt = "%5.1f"

        # If the user passed --integer
        if op.integer:
            self.num_fmt = "%5.0f"

        self.open('/proc/meminfo')

    def colorize_percent(self, num, red_high):

        if red_high:
            color_low = theme['colors_lo'][0]
            color_med = theme['colors_lo'][1]
            color_hig = theme['colors_lo'][2]
        else:
            color_low = theme['colors_lo'][2]
            color_med = theme['colors_lo'][1]
            color_hig = theme['colors_lo'][0]

        num_str = self.num_fmt % (num)

        # We're doing a simple red/yellow/green every 33%
        if num <= 33.3333:
            ret = color_low + num_str + ansi['reset']
        elif num <= 66.6666:
            ret = color_med + num_str + ansi['reset']
        else:
            ret = color_hig + num_str + ansi['reset']

        return ret

    def extract(self):
        extra_extract_vars = ('MemTotal', 'Shmem', 'SReclaimable')
        extra_val          = {}

        for l in self.splitlines():
            if len(l) < 2: continue
            name = l[0].split(':')[0]

            if name in self.vars:
                self.val[name] = int(l[1]) * 1024.0
            if name in extra_extract_vars:
                extra_val[name] = int(l[1]) * 1024.0

        total  = extra_val['MemTotal']
        free   = (self.val['MemFree']      / total) * 100
        avail  = (self.val['MemAvailable'] / total) * 100
        cached = (self.val['Cached']       / total) * 100

        # This math is borrow from the regular mem plugin
        used   = (((extra_val['MemTotal'] - self.val['Cached'] - extra_val['SReclaimable']) / total) * 100)

        self.val['MemUsed']      = self.colorize_percent(used, False)
        self.val['MemFree']      = self.colorize_percent(free, True)
        self.val['MemAvailable'] = self.colorize_percent(avail, True)

        # Cached is NOT colorized
        self.val['Cached'] = self.num_fmt % cached





### Author: Dag Wieers <dag$wieers,com>

### FIXME: This module needs infrastructure to provide a list of mountpoints
### FIXME: Would be nice to have a total by default (half implemented)

class dool_plugin(dool):
    """
    Amount of used and free space per mountpoint.
    """

    def __init__(self):
        self.nick = ('used', 'free')
        self.open('/etc/mtab')
        self.cols = 2

    def vars(self):
        ret = []

        mystr = os.environ.get('DOOL_FREESPACE_MOUNT_POINTS','').strip()

        if (len(mystr) > 0):
            mp = mystr.split(',')
        else:
            mp = []

        include_fs_types = (
            'ext2', 'ext3', 'ext4', 'btrfs', 'xfs'
        )

        for l in self.splitlines():
            if len(l) < 6: continue

            device      = l[0]
            mount_point = l[1]
            fs_type     = l[2]

            #print(device + " | " + mount_point + " | " + fs_type)

            # If there is an array of mount points (whitelist) and this
            # mount point is *NOT* in that list, skip it
            if (mp):
                if (not mount_point in mp):
                    continue
            # If there is NOT an array of mount points check it against a
            # whitelisted array of fs_types
            elif (fs_type not in include_fs_types):
                continue

            res = os.statvfs(mount_point)

            if res[0] == 0: continue ### Skip zero block filesystems
            ret.append(mount_point)

        return ret

    def name(self):
        return ['/' + os.path.basename(name) for name in self.vars]

    def extract(self):
        self.val['total'] = (0, 0)
        for name in self.vars:
            res = os.statvfs(name)
            self.val[name] = ( (float(res.f_blocks) - float(res.f_bavail)) * int(res.f_frsize), float(res.f_bavail) * float(res.f_frsize) )
            self.val['total'] = (self.val['total'][0] + self.val[name][0], self.val['total'][1] + self.val[name][1])

# vim:ts=4:sw=4:et



### Author: Dag Wieers <dag$wieers,com>

class dool_plugin(dool):
    """
    Percentage of bandwidth utilization for block devices.

    Displays percentage of CPU time during which I/O requests were issued
    to the device (bandwidth utilization for the device). Device saturation
    occurs when this value is close to 100%.
    """

    def __init__(self):
        self.nick   = ('util', )
        self.type   = 'f'
        self.width  = 4
        self.scale  = 34
        self.cols   = 1
        self.struct = dict( tot_ticks=0 )
        self.open('/proc/diskstats')

    def discover(self, *objlist):
        ret = []
        for l in self.splitlines():
            if len(l) < 13: continue
            if set(l[3:]) == {'0'}: continue
            name = l[2]
            ret.append(name)
        for item in objlist: ret.append(item)
        if not ret:
            raise Exception('No suitable block devices found to monitor')

        return ret

    def basename(self, disk):
        "Strip /dev/ and convert symbolic link"
        if disk[:5] == '/dev/':
            # file or symlink
            if os.path.exists(disk):
                # e.g. /dev/disk/by-uuid/15e40cc5-85de-40ea-b8fb-cb3a2eaf872
                if os.path.islink(disk):
                    target = os.readlink(disk)
                    # convert relative pathname to absolute
                    if target[0] != '/':
                        target = os.path.join(os.path.dirname(disk), target)
                        target = os.path.normpath(target)
                    print('dstat: symlink %s -> %s' % (disk, target))
                    disk = target
                # trim leading /dev/
                return disk[5:]
            else:
                print('dstat: %s does not exist' % disk)
        else:
            return disk

    def vars(self):
        ret     = []
        varlist = []

        # If there was a disk filter specified on the CLI
        # example: dool --disk-infligh -D md123,sda
        if op.disklist:
            for x in op.disklist:
                base = self.basename(x)
                varlist.append(base)
        # We're grabbing them all on the fly via discover()
        else:
            for name in self.discover:
                if DOOL_DISKFILTER.match(name): continue
                if name not in blockdevices(): continue
                varlist.append(name)

            varlist.sort()

        for name in varlist:
            if name in self.discover:
                ret.append(name)

        return ret

    def name(self):
        ret = []
        for x in self.vars:
            ret.append(dev_short_name(x))

        return ret

    def extract(self):
        for l in self.splitlines():
            if len(l) < 13: continue
            if l[5] == '0' and l[9] == '0': continue
            if set(l[3:]) == {'0'}: continue
            name = l[2]
            if name not in self.vars: continue
            self.set2[name] = dict(
                tot_ticks = int(l[12])
            )

        for name in self.vars:
            self.val[name] = ( (self.set2[name]['tot_ticks'] - self.set1[name]['tot_ticks']) * 1.0 * hz / elapsed / 1000, )

        if step == op.delay:
            self.set1.update(self.set2)

# vim: tabstop=4 shiftwidth=4 expandtab autoindent softtabstop=4



# Author: Brock Palen <brockp@mlds-networks.com>, Kilian Cavalotti <kilian@stanford.edu>

class dool_plugin(dool):
    def __init__(self):
        self.nick = ('read', 'write')
        self.cols = 2
        self.stat_path = '/proc/fs/lustre/llite'

    def check(self):
        if os.path.exists('/proc/fs/lustre/llite'):
            self.stat_path = '/proc/fs/lustre/llite'
        elif os.path.exists('/sys/kernel/debug/lustre/llite'):
            self.stat_path = '/sys/kernel/debug/lustre/llite'
        else:
            raise Exception('Lustre filesystem not found')
        info(1, 'Module %s is still experimental.' % self.filename)

    def name(self):
        return [mount for mount in os.listdir(self.stat_path)]

    def vars(self):
        return [mount for mount in os.listdir(self.stat_path)]

    def extract(self):
        for name in self.vars:
            read = write = 0
            for line in dopen(os.path.join(self.stat_path, name, 'stats')).readlines():
                l = line.split()
                if len(l) < 6: continue
                if l[0] == 'read_bytes':
                    read = int(l[6])
                elif l[0] == 'write_bytes':
                    write = int(l[6])
            self.set2[name] = (read, write)

            self.val[name] = list(map(lambda x, y: (y - x) * 1.0 / elapsed, self.set1[name], self.set2[name]))

        if step == op.delay:
            self.set1.update(self.set2)

# vim:ts=4:sw=4



### Author: Dag Wieers <dag$wieers,com>

class dool_plugin(dool):
    def __init__(self):
        self.name = 'memory usage'
        self.nick = ('used', 'buff', 'cach', 'free')
        self.vars = ('MemUsed', 'Buffers', 'Cached', 'MemFree')
        self.server = os.getenv('DOOL_SNMPSERVER') or '192.168.1.1'
        self.community = os.getenv('DOOL_SNMPCOMMUNITY') or 'public'

    def check(self):
        try:
            global cmdgen
            from pysnmp.entity.rfc3413.oneliner import cmdgen
        except:
            raise Exception('Needs pysnmp and pyasn1 modules')

    def extract(self):
        self.val['MemTotal'] = int(snmpget(self.server, self.community, (1,3,6,1,4,1,2021,4,5,0))) * 1024
        self.val['MemFree'] = int(snmpget(self.server, self.community, (1,3,6,1,4,1,2021,4,11,0))) * 1024
#        self.val['Shared'] = int(snmpget(self.server, self.community, (1,3,6,1,4,1,2021,4,13,0))) * 1024
        self.val['Buffers'] = int(snmpget(self.server, self.community, (1,3,6,1,4,1,2021,4,14,0))) * 1024
        self.val['Cached'] = int(snmpget(self.server, self.community, (1,3,6,1,4,1,2021,4,15,0))) * 1024

        self.val['MemUsed'] = self.val['MemTotal'] - self.val['MemFree']

# vim:ts=4:sw=4:et



### Author: <krikava$gmail,com>

### Condor queue plugin
### Display information about jobs in queue (using condor_q(1))
###
### WARNING: with many jobs in the queue, the condor_q might take quite
### some time to execute and use quite a bit of resources. Consider
### using a longer delay.

import os
import re

global condor_classad

class condor_classad:
    """
    Utility class to work with Condor ClassAds
    """

    global ATTR_VAR_PATTERN
    ATTR_VAR_PATTERN = re.compile(r'\$\((\w+)\)')

    def __init__(self, file=None, config=None):
        if file != None:
            self.attributes = condor_classad._read_from_file(file)
        elif config != None:
            self.attributes = condor_classad._parse(config)

        if self.attributes == None:
            raise Exception('condor_config must be initialized either using a file or config text')

        local_config_file = self['LOCAL_CONFIG_FILE']

        if local_config_file != None:
            for k,v in condor_classad._read_from_file(local_config_file).items():
                self.attributes[k] = v

    def __getitem__(self, name):
        if name in self.attributes:
            self._expand(name)
        return self.attributes[name]

    def _expand(self, var):
        if not var in self.attributes:
            return

        while True:
            m = ATTR_VAR_PATTERN.match(self.attributes[var])
            if m == None:
                break
            var_name = m.group(1)
            self.attributes[var] = ATTR_VAR_PATTERN.sub(self.attributes[var_name],
                                                          self.attributes[var])

    @staticmethod
    def _parse(text):
        attributes = {}
        for l in [l for l in text.split('\n') if not l.strip().startswith('#')]:
            l = l.split('=')
            if len(l) <= 1 or len(l[0]) == 0:
                continue
            attributes[l[0].strip()] = ''.join(l[1:]).strip()
        return attributes

    @staticmethod
    def _read_from_file(filename):
        if not os.access(filename, os.R_OK):
            raise Exception('Unable to read file %s' % filename)
        try:
            f = open(filename)
            return condor_classad._parse((f.read()))
        finally:
            f.close()

class dool_plugin(dool):
    """
    Plugin for Condor queue stats
    """

    global CONDOR_Q_STAT_PATTER
    CONDOR_Q_STAT_PATTER = re.compile(r'(\d+) jobs; (\d+) idle, (\d+) running, (\d+) held')

    def __init__(self):
        self.name = 'condor queue'
        self.vars = ('jobs', 'idle', 'running', 'held')
        self.type = 'd'
        self.width = 5
        self.scale = 1
        self.condor_config = None

    def check(self):
        config_file = os.environ['CONDOR_CONFIG']
        if config_file == None:
            raise Exception('Environment varibale CONDOR_CONFIG is missing')
        self.condor_config = condor_classad(config_file)

        bin_dir = self.condor_config['BIN']
        if bin_dir == None:
            raise Exception('Unable to find BIN directory in condor config file %s' % config_file)

        self.condor_status_cmd = os.path.join(bin_dir, 'condor_q')

        if not os.access(self.condor_status_cmd, os.X_OK):
            raise Exception('Needs %s in the path' % self.condor_status_cmd)
        else:
            try:
                p = os.popen(self.condor_status_cmd+' 2>&1 /dev/null')
                ret = p.close()
                if ret:
                    raise Exception('Cannot interface with Condor - condor_q returned != 0?')
            except IOError:
                raise Exception('Unable to execute %s' % self.condor_status_cmd)
            return True

    def extract(self):
        last_line = None

        try:
            for repeats in range(3):
                for last_line in cmd_readlines(self.condor_status_cmd):
                    pass

                m = CONDOR_Q_STAT_PATTER.match(last_line)
                if m == None:
                    raise Exception('Invalid output from %s. Got: %s' % (cmd, last_line))

                stats = [int(s.strip()) for s in m.groups()]
                for i,j in enumerate(self.vars):
                    self.val[j] = stats[i]
        except Exception:
            for name in self.vars:
                self.val[name] = -1

# vim:ts=4:sw=4:et



### Author: Dag Wieers <dag$wieers,com>

class dool_plugin(dool):
    """
    Provide more information related to the dstat process.

    The dstat cputime is the total cputime dstat requires per second. On a
    system with one cpu and one core, the total cputime is 1000ms. On a system
    with 2 cores the total is 2000ms. It may help to vizualise the performance
    of Dstat and its selection of plugins.
    """
    def __init__(self):
        self.name = 'dstat'
        self.vars = ('cputime', 'latency')
        self.type = 'd'
        self.width = 5
        self.scale = 1000
        self.open('/proc/%s/schedstat' % ownpid)

    def extract(self):
        l = self.splitline()
#        l = linecache.getline('/proc/%s/schedstat' % self.pid, 1).split()
        self.set2['cputime'] = int(l[0])
        self.set2['latency'] = int(l[1])

        for name in self.vars:
            self.val[name] = (self.set2[name] - self.set1[name]) * 1.0 / elapsed

        if step == op.delay:
            self.set1.update(self.set2)

# vim:ts=4:sw=4:et



### Author: Bert de Bruijn <bert+dstat$debruijn,be>

### VMware ESX kernel interrupt stats
### Displays kernel interrupt statistics on VMware ESX servers

# NOTE TO USERS: command-line plugin configuration is not yet possible, so I've
# "borrowed" the -I argument. 
# EXAMPLES:
# # dstat --vmkint -I 0x46,0x5a
# You can even combine the Linux and VMkernel interrupt stats
# # dstat --vmkint -i -I 14,0x5a
# Look at /proc/vmware/interrupts to see which interrupt is linked to which function

class dool_plugin(dool):
    def __init__(self):
        self.name = 'vmkint'
        self.type = 'd'
        self.width = 4
        self.scale = 1000
        self.open('/proc/vmware/interrupts')
#       self.intmap = self.intmap()

#   def intmap(self):
#       ret = {}
#       for line in dopen('/proc/vmware/interrupts').readlines():
#           l = line.split()
#           if len(l) <= self.vmkcpunr: continue
#           l1 = l[0].split(':')[0]
#           l2 = ' '.join(l[vmkcpunr()+1:]).split(',')
#           ret[l1] = l1
#           for name in l2:
#               ret[name.strip().lower()] = l1
#           return ret

    def vmkcpunr(self):
        #the service console sees only one CPU, so cpunr == 1, only the vmkernel sees all CPUs
        ret = []
        # default cpu number is 2
        ret = 2
        for l in self.fd[0].splitlines():
            if l[0] == 'Vector': 
                ret = int( int( l[-1] ) + 1 )
        return ret

    def discover(self):
        #interrupt names are not decimal numbers, but rather hexadecimal numbers like 0x7e
        ret = []
        self.fd[0].seek(0)
        for line in self.fd[0].readlines():
            l = line.split()
            if l[0] == 'Vector': continue
            if len(l) < self.vmkcpunr()+1: continue
            name = l[0].split(':')[0]
            amount = 0
            for i in l[1:1+self.vmkcpunr()]:
                amount = amount + int(i)
            if amount > 20: ret.append(str(name))
        return ret

    def vars(self):
        ret = []
        if op.intlist:
            list = op.intlist
        else:
            list = self.discover
#           len(list) > 5: list = list[-5:]
        for name in list:
            if name in self.discover:
                ret.append(name)
#           elif name.lower() in self.intmap:
#               ret.append(self.intmap[name.lower()])
        return ret

    def check(self): 
        try:
            os.listdir('/proc/vmware')
        except:
            raise Exception('Needs VMware ESX')
        info(1, 'The vmkint module is an EXPERIMENTAL module.')

    def extract(self):
        self.fd[0].seek(0)
        for line in self.fd[0].readlines():
            l = line.split()
            if len(l) < self.vmkcpunr()+1: continue
            name = l[0].split(':')[0]
            if name in self.vars:
                self.set2[name] = 0
                for i in l[1:1+self.vmkcpunr()]:
                    self.set2[name] = self.set2[name] + int(i)

        for name in self.set2:
            self.val[name] = (self.set2[name] - self.set1[name]) * 1.0 / elapsed

        if step == op.delay:
            self.set1.update(self.set2)

# vim:ts=4:sw=4



### Author: Dag Wieers <dag$wieers,com>

class dool_plugin(dool):
    """
    The average queue length of the requests that were issued to the device.
    """

    def __init__(self):
        self.version = 2
        self.nick = ('avgqu',)
        self.type = 'f'
        self.width = 4
        self.scale = 10
        self.open('/proc/diskstats')
        self.cols = 1
        self.struct = dict( rq_ticks=0 )

    def discover(self, *objlist):
        ret = []
        for l in self.splitlines():
            if len(l) < 13: continue
            if set(l[3:]) == {'0'}: continue
            name = l[2]
            ret.append(name)
        for item in objlist: ret.append(item)
        if not ret:
            raise Exception('No suitable block devices found to monitor')
        return ret

    def vars(self):
        ret = []
        if op.disklist:
            varlist = op.disklist
        else:
            varlist = []
            blockdevices = [os.path.basename(filename) for filename in glob.glob('/sys/block/*')]
            for name in self.discover:
                if DOOL_DISKFILTER.match(name): continue
                if name not in blockdevices: continue
                varlist.append(name)
            varlist.sort()
        for name in varlist:
            if name in self.discover:
                ret.append(name)
        return ret

    def name(self):
        return self.vars

    def extract(self):
        for l in self.splitlines():
            if len(l) < 13: continue
            if set(l[3:]) == {'0'}: continue
            if l[3] == '0' and l[7] == '0': continue
            name = l[2]
            if name not in self.vars or name == 'total': continue
            self.set2[name] = dict(
                rq_ticks = int(l[13]),
            )

        for name in self.vars:
            self.val[name] = ( ( self.set2[name]['rq_ticks'] - self.set1[name]['rq_ticks'] ) * 1.0 / elapsed / 1000, )

        if step == op.delay:
            self.set1.update(self.set2)



### Author: Scott Baker <scott@perturb.org>

class dool_plugin(dool):
	"""
	The number of in-flight IO operations pending per disk
	"""

	def __init__(self):
		# Each column is X chars wide
		self.width = 4 # Defaults to 5 if not specified

		devices    = self.get_devs()
		short_devs = []
		for x in devices:
			short_name = dev_short_name(x, self.width)
			short_devs.append(short_name)

		# This is the plugin name that goes on top of the columns
		#self.name	= 'io/inflight'
		self.name  = short_devs
		# Title of each column (or group of colums)
		self.nick = ('ioif',)
		# These are the sub headings under the main column (each disk device)
		self.vars  = devices
		# It's a decimal number
		self.type  = 'd'
		# Group the colorings to chunks of 20
		self.scale = 20
		# The number of columns we're showing
		self.cols  = len(devices)

	def get_devs(self):
		# If there was a disk filter specified on the CLI
		# example: dool --disk-infligh -D md123,sda
		if op.disklist:
			return op.disklist

		ret   = []
		globx = glob.glob('/sys/block/*')

		for path in globx:
			base = os.path.basename(path)
			ret.append(base)

		ret.sort()

		return ret

	def extract(self):
		for dev in self.vars:
			# Documentation for /sys/block/$DEV/stat
			# https://www.kernel.org/doc/Documentation/block/stat.txt
			path = ("/sys/block/%s/stat" % (dev))

			for l in proc_splitlines(path):
				value = int(l[8])

				# For testing we can toss random data in the buckets
				#value = random.randrange(1,100)

				self.val[dev] = (value,)

# vim: tabstop=4 shiftwidth=4 noexpandtab autoindent softtabstop=4



### Author: Bert de Bruijn <bert+dstat$debruijn,be>

### VMware memory stats
### Displays memory stats coming from the hypervisor inside VMware VMs.
### The vmGuestLib API from VMware Tools needs to be installed

class dool_plugin(dool):
    def __init__(self):
        self.name = 'vmware memory'
        self.vars = ('active', 'ballooned', 'mapped',  'swapped', 'used')
        self.nick = ('active', 'balln', 'mappd', 'swapd', 'used')
        self.type = 'd'
        self.width = 5
        self.scale = 1024

    def check(self):
        try:
            global vmguestlib
            import vmguestlib

            self.gl = vmguestlib.VMGuestLib()
        except:
            raise Exception('Needs python-vmguestlib module')

    def extract(self):
        self.gl.UpdateInfo()
        self.val['active'] = self.gl.GetMemActiveMB() * 1024 ** 2
        self.val['ballooned'] = self.gl.GetMemBalloonedMB() * 1024 ** 2
        self.val['mapped'] = self.gl.GetMemMappedMB() * 1024 ** 2
        self.val['swapped'] = self.gl.GetMemSwappedMB() * 1024 ** 2
        self.val['used'] = self.gl.GetMemUsedMB() * 1024 ** 2

# vim:ts=4:sw=4



### Author: Dag Wieers <dag$wieers,com>

class dool_plugin(dool):
    """
    Remaining battery time.

    Calculated from power drain and remaining battery power. Information is
    retrieved from ACPI.
    """

    def __init__(self):
        self.name = 'remain'
        self.type = 't'
        self.width = 5
        self.scale = 0

    def vars(self):
        ret = []
        for battery in os.listdir('/proc/acpi/battery/'):
            for line in dopen('/proc/acpi/battery/'+battery+'/state').readlines():
                l = line.split()
                if len(l) < 2: continue
                if l[0] == 'present:' and l[1] == 'yes':
                    ret.append(battery)
        ret.sort()
        return ret

    def nick(self):
        return [name.lower() for name in self.vars]

    def extract(self):
        for battery in self.vars:
            for line in dopen('/proc/acpi/battery/'+battery+'/state').readlines():
                l = line.split()
                if len(l) < 3: continue
                if l[0:2] == ['remaining', 'capacity:']:
                    remaining = int(l[2])
                    continue
                elif l[0:2] == ['present', 'rate:']:
                    rate = int(l[2])
                    continue

            if rate and remaining:
                self.val[battery] = remaining * 60 / rate
            else:
                self.val[battery] = -1

# vim:ts=4:sw=4:et



### Author: <gianfranco@mongodb.com>

global mongodb_user
mongodb_user = os.getenv('DOOL_MONGODB_USER') or os.getenv('USER')

global mongodb_pwd
mongodb_pwd = os.getenv('DOOL_MONGODB_PWD')

global mongodb_host
mongodb_host = os.getenv('DOOL_MONGODB_HOST') or '127.0.0.1:27017'

class dool_plugin(dool):
  """
  Plugin for MongoDB.
  """
  def __init__(self):
    global pymongo
    import pymongo

    try:
      self.m = pymongo.MongoClient(mongodb_host)
      if mongodb_pwd:
        self.m.admin.authenticate(mongodb_user, mongodb_pwd)
      self.db = self.m.admin
    except Exception as e:
      raise Exception('Cannot interface with MongoDB server: %s' % e)

    self.name    = 'mongodb counts'
    self.nick    = ('qry', 'ins', 'upd', 'del', 'gtm', 'cmd')
    self.vars    = ('query', 'insert','update','delete','getmore','command')
    self.type    = 'd'
    self.width   = 5
    self.scale   = 2
    self.lastVal = {}

  def extract(self):
    status = self.db.command("serverStatus")
    opct = status['opcounters']

    for name in self.vars:
      if name in opct.iterkeys():
        if not name in self.lastVal:
          self.lastVal[name] = opct.get(name)

        self.val[name]     = (int(opct.get(name)) - self.lastVal[name]) / elapsed
        self.lastVal[name] = opct.get(name)



### Author: Bert de Bruijn <bert+dstat$debruijn,be>

### VMware ESX kernel vmknic stats
### Displays VMkernel port statistics on VMware ESX servers

# NOTE TO USERS: command-line plugin configuration is not yet possible, so I've
# "borrowed" the -N argument.
# EXAMPLES:
# # dstat --vmknic -N vmk1
# You can even combine the Linux and VMkernel network stats (just don't just "total").
# # dstat --vmknic -n -N vmk0,vswif0
# NB Data comes from /proc/vmware/net/tcpip/ifconfig

class dool_plugin(dool):
    def __init__(self):
        self.name = 'vmknic'
        self.nick = ('recv', 'send')
        self.open('/proc/vmware/net/tcpip/ifconfig')
        self.cols = 2

    def check(self):
        try:
            os.listdir('/proc/vmware')
        except:
            raise Exception('Needs VMware ESX')
        info(1, 'The vmknic module is an EXPERIMENTAL module.')

    def discover(self, *list):
        ret = []
        for l in self.fd[0].splitlines(replace=' /', delim='/'):
            if len(l) != 12: continue
            if l[2][:5] == '<Link': continue
            if ','.join(l) == 'Name,Mtu/TSO,Network,Address,Ipkts,Ierrs,Ibytes,Opkts,Oerrs,Obytes,Coll,Time': continue
            if l[0] == 'lo0': continue
            if l[0] == 'Usage:': continue
            ret.append(l[0])
        ret.sort()
        for item in list: ret.append(item)
        return ret

    def vars(self):
        ret = []
        if op.netlist:
            list = op.netlist
        else:
            list = self.discover
            list.sort()
        for name in list:
            if name in self.discover + ['total']:
                ret.append(name)
        return ret

    def name(self):
        return ['net/'+name for name in self.vars]

    def extract(self):
        self.set2['total'] = [0, 0]
        for line in self.fd[0].readlines():
            l = line.replace(' /','/').split()
            if len(l) != 12: continue
            if l[2][:5] == '<Link': continue
            if ','.join(l) == 'Name,Mtu/TSO,Network,Address,Ipkts,Ierrs,Ibytes,Opkts,Oerrs,Obytes,Coll,Time': continue
            if l[0] == 'Usage:': continue
            name = l[0]
            if name in self.vars:
                self.set2[name] = ( int(l[6]), int(l[9]) )
            if name != 'lo0':
                self.set2['total'] = ( self.set2['total'][0] + int(l[6]), self.set2['total'][1] + int(l[9]) )

        if update:
            for name in self.set2:
                self.val[name] = list(map(lambda x, y: (y - x) * 1.0 / elapsed, self.set1[name], self.set2[name]))

        if step == op.delay:
            self.set1.update(self.set2)

# vim:ts=4:sw=4



### Author: <gianfranco@mongodb.com>

global mongodb_user
mongodb_user = os.getenv('DOOL_MONGODB_USER') or os.getenv('USER')

global mongodb_pwd
mongodb_pwd = os.getenv('DOOL_MONGODB_PWD')

global mongodb_host
mongodb_host = os.getenv('DOOL_MONGODB_HOST') or '127.0.0.1:27017'

class dool_plugin(dool):
  """
  Plugin for MongoDB.
  """
  def __init__(self):
    global pymongo
    import pymongo

    try:
      self.m = pymongo.MongoClient(mongodb_host)
      if mongodb_pwd:
        self.m.admin.authenticate(mongodb_user, mongodb_pwd)
      self.db = self.m.admin
    except Exception as e:
      raise Exception('Cannot interface with MongoDB server: %s' % e)

    self.name    = 'mongodb queues'
    self.nick    = ('ar', 'aw', 'qt', 'qw')
    self.vars    = ('ar', 'aw', 'qt', 'qw')
    self.type    = 'd'
    self.width   = 5
    self.scale   = 2
    self.lastVal = {}

  def extract(self):
    status = self.db.command("serverStatus")
    glock = status['globalLock']
    alock = glock['activeClients']
    qlock = glock['currentQueue']

    self.val['ar'] = int(alock['readers'])
    self.val['aw'] = int(alock['writers'])
    self.val['qr'] = int(qlock['readers'])
    self.val['qw'] = int(qlock['writers'])



### Dstat all I/O process plugin
### Displays all processes' I/O read/write stats and CPU usage
###
### Authority: Guillermo Cantu Luna

class dool_plugin(dool):
	def __init__(self):
		self.name    = 'most expensive i/o process'
		self.vars    = ('process               pid    read  writ  cpu ',)
		self.type    = 's'
		self.width   = 45
		self.scale   = 0
		self.pidset1 = {}

	def check(self):
		if not os.access('/proc/self/io', os.R_OK):
			raise Exception('Kernel has no per-process I/O accounting [CONFIG_TASK_IO_ACCOUNTING], use at least 2.6.20')
		return True

	def extract(self):
		self.output = ''
		self.pidset2 = {}
		self.val['usage'] = 0.0
		for pid in proc_pidlist():
			try:
				### Reset values
				if pid not in self.pidset2:
					self.pidset2[pid] = {'rchar:': 0, 'wchar:': 0, 'cputime:': 0, 'cpuper:': 0}
				if pid not in self.pidset1:
					self.pidset1[pid] = {'rchar:': 0, 'wchar:': 0, 'cputime:': 0, 'cpuper:': 0}

				### Extract name
				name = proc_splitline('/proc/%s/stat' % pid)[1][1:-1]

				### Extract counters
				for l in proc_splitlines('/proc/%s/io' % pid):
					if len(l) != 2: continue
					self.pidset2[pid][l[0]] = int(l[1])

				### Get CPU usage
				l = proc_splitline('/proc/%s/stat' % pid)
				if len(l) < 15:
					cpu_usage = 0
				else:
					self.pidset2[pid]['cputime:'] = int(l[13]) + int(l[14])
					cpu_usage = (self.pidset2[pid]['cputime:'] - self.pidset1[pid]['cputime:']) * 1.0 / elapsed / cpunr

			except ValueError:
				continue
			except IOError:
				continue
			except IndexError:
				continue

			read_usage  = (self.pidset2[pid]['rchar:'] - self.pidset1[pid]['rchar:']) * 1.0 / elapsed
			write_usage = (self.pidset2[pid]['wchar:'] - self.pidset1[pid]['wchar:']) * 1.0 / elapsed
			usage       = read_usage + write_usage

			### Get the process that spends the most jiffies
			if usage > self.val['usage']:
				self.val['usage']       = usage
				self.val['read_usage']  = read_usage
				self.val['write_usage'] = write_usage
				self.val['pid']         = pid
				self.val['name']        = getnamebypid(pid, name)
				self.val['cpu_usage']   = cpu_usage

		if step == op.delay:
			self.pidset1 = self.pidset2

		if self.val['usage'] != 0.0:
			#self.output = '%-*s%s%-5s%s%s%s%s%%' % (self.width-14-len(pid), self.val['name'][0:self.width-14-len(pid)], color['darkblue'], self.val['pid'], cprint(self.val['read_usage'], 'd', 5, 1024), cprint(self.val['write_usage'], 'd', 5, 1024), cprint(self.val['cpu_usage'], 'f', 3, 34), color['darkgray'])

			pid_str   = color['darkblue'] + ("%-6s" % self.val['pid']) + ansi['reset']
			read_str  = cprint(self.val['read_usage'] , 'd', 5, 1024)
			write_str = cprint(self.val['write_usage'], 'd', 5, 1024)
			cpu_str   = cprint(self.val['cpu_usage']  , 'f', 4, 34)

			self.output = '%-20s %s %s %s %s%%' % (self.val['name'], pid_str, read_str, write_str, cpu_str)

	def showcsv(self):
		return 'Top: %s\t%s\t%s\t%s' % (self.val['name'][0:self.width-20], self.val['read_usage'], self.val['write_usage'], self.val['cpu_usage'])

# vim: tabstop=4 shiftwidth=4 noexpandtab autoindent softtabstop=4



### Author: Adam Michel <elfurbe@furbism.com>
### Based on work by: Dag Wieers <dag@wieers.com>

class dool_plugin(dool):
    def __init__(self):
        self.name = 'nfs4 client'
        # this vars/nick pair is the ones I considered relevant. Any set of the full list would work.
        self.vars = ('read', 'write', 'readdir', 'commit', 'getattr', 'create', 'link','remove')
        self.nick = ('read', 'writ', 'rdir', 'cmmt', 'gatr','crt','link','rmv')
        # this is every possible variable if you're into that
        #self.vars = ("read", "write", "commit", "open", "open_conf", "open_noat", "open_dgrd", "close", 
        #        "setattr", "fsinfo", "renew", "setclntid", "confirm", "lock", "lockt", "locku", 
        #        "access", "getattr", "lookup", "lookup_root", "remove", "rename", "link", "symlink", 
        #        "create", "pathconf", "statfs", "readlink", "readdir", "server_caps", "delegreturn", 
        #        "getacl", "setacl", "fs_locations", "rel_lkowner", "secinfo")
        # these are terrible shortnames for every possible variable
        #self.nick = ("read", "writ", "comt", "open", "opnc", "opnn", "opnd", "clse", "seta", "fnfo", 
        #        "renw", "stcd", "cnfm", "lock", "lckt", "lcku", "accs", "gatr", "lkup", "lkp_r", 
        #        "rem", "ren", "lnk", "slnk", "crte", "pthc", "stfs", "rdlk", "rdir", "scps", "delr", 
        #        "gacl", "sacl", "fslo", "relo", "seco")
        self.type = 'd'
        self.width = 5
        self.scale = 1000
        self.open('/proc/net/rpc/nfs')

    def check(self):
        # other NFS modules had this, so I left it. It seems to work.
        info(1, 'Module %s is still experimental.' % self.filename)

    def extract(self):
        # list of fields from nfsstat, in order of output from cat /proc/net/rpc/nfs 
        nfs4_names = ("version", "fieldcount", "null", "read", "write", "commit", "open", "open_conf",
                "open_noat", "open_dgrd", "close", "setattr", "fsinfo", "renew", "setclntid",
                "confirm", "lock", "lockt", "locku", "access", "getattr", "lookup", "lookup_root",
                "remove", "rename", "link", "symlink", "create", "pathconf", "statfs", "readlink",
                "readdir", "server_caps", "delegreturn", "getacl", "setacl", "fs_locations",
                "rel_lkowner", "secinfo")
        for line in self.splitlines():

            fields = line.split()
            if fields[0] == "proc4": # just grab NFSv4 stats
                assert int(fields[1]) == len(fields[2:]), ("reported field count (%d) does not match actual field count (%d)" % (int(fields[1]), len(fields[2:])))
                for var in self.vars:
                    self.set2[var] = fields[nfs4_names.index(var)]

        for name in self.vars:
            self.val[name] = (int(self.set2[name]) - int(self.set1[name])) * 1.0 / elapsed

        if step == op.delay:
            self.set1.update(self.set2)

# vim:ts=4:sw=4:et



### Author: <lefred$inuits,be>

global mysql_user
mysql_user = os.getenv('DOOL_MYSQL_USER') or os.getenv('USER')

global mysql_pwd
mysql_pwd = os.getenv('DOOL_MYSQL_PWD')

global mysql_host
mysql_host = os.getenv('DOOL_MYSQL_HOST')

global mysql_port
mysql_port = os.getenv('DOOL_MYSQL_PORT')

global mysql_socket
mysql_socket = os.getenv('DOOL_MYSQL_SOCKET')

class dool_plugin(dool):
    """
    Plugin for MySQL 5 connections.
    """

    def __init__(self):
        self.name = 'mysql5 conn'
        self.nick = ('ThCon', '%Con')
        self.vars = ('Threads_connected', 'Threads')
        self.type = 'f'
        self.width = 4
        self.scale = 1

    def check(self): 
        global MySQLdb
        import MySQLdb
        try:
            args = {}
            if mysql_user:
                args['user'] = mysql_user
            if mysql_pwd:
                args['passwd'] = mysql_pwd
            if mysql_host:
                args['host'] = mysql_host
            if mysql_port:
                args['port'] = mysql_port
            if mysql_socket:
                args['unix_socket'] = mysql_socket

            self.db = MySQLdb.connect(**args)
        except Exception as e:
            raise Exception('Cannot interface with MySQL server, %s' % e)

    def extract(self):
        try:
            c = self.db.cursor()
            c.execute("""show global variables like 'max_connections';""")
            max = c.fetchone()
            c.execute("""show global status like 'Threads_connected';""")
            thread = c.fetchone()
            if thread[0] in self.vars:
                self.set2[thread[0]] = float(thread[1])
                self.set2['Threads'] = float(thread[1]) / float(max[1]) * 100.0

            for name in self.vars:
                self.val[name] = self.set2[name] * 1.0 / elapsed

            if step == op.delay:
                self.set1.update(self.set2)

        except Exception as e:
            for name in self.vars:
                self.val[name] = -1

# vim:ts=4:sw=4:et



### Author: Jihyun Yu <yjh0502@gmail.com>

global redis_host 
redis_host = os.getenv('DOOL_REDIS_HOST') or "127.0.0.1"

global redis_port
redis_port = os.getenv('DOOL_REDIS_PORT') or "6379"

class dool_plugin(dool):
    def __init__(self):
        self.type = 'd'
        self.width = 7
        self.scale = 10000
        self.name = 'redis'
        self.nick = ('tps',)
        self.vars = ('tps',)
        self.cmdInfo = '*1\r\n$4\r\ninfo\r\n'

    def get_info(self):
        global socket
        import socket

        global redis_host
        global redis_port

        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.settimeout(0.1)
            s.connect((redis_host, int(redis_port)))
            s.send(self.cmdInfo)
            dict = {};
            for line in s.recv(1024*1024).split('\r\n'):
                if line == "" or line[0] == '#' or line[0] == '*' or line[0] == '$':
                    continue
                pair = line.split(':', 2)
                dict[pair[0]] = pair[1]
            return dict
        except:
            return {}
        finally:
            try:
                s.close()
            except:
                pass

    def extract(self):
        key = "instantaneous_ops_per_sec"
        dic = self.get_info()
        if key in dic:
            self.val['tps'] = int(dic[key])

# vim:ts=4:sw=4:et



### Author: Dag Wieers <dag@wieers.com>

class dool_plugin(dool):
    """
    Number of packets received and send per interface.
    """

    def __init__(self):
        self.nick = ('#recv', '#send')
        self.type = 'd'
        self.width = 5
        self.scale = 1000
        self.totalfilter = re.compile(r'^(lo|bond\d+|face|.+\.\d+)$')
        self.open('/proc/net/dev')
        self.cols = 2

    def discover(self, *objlist):
        ret = []
        for l in self.splitlines(replace=':'):
            if len(l) < 17: continue
            if l[2] == '0' and l[10] == '0': continue
            name = l[0]
            if name not in ('lo', 'face'):
                ret.append(name)
        ret.sort()
        for item in objlist: ret.append(item)
        return ret

    def vars(self):
        ret = []
        if op.netlist:
            varlist = op.netlist
        elif not op.full:
            varlist = ('total',)
        else:
            varlist = self.discover
#           if len(varlist) > 2: varlist = varlist[0:2]
            varlist.sort()
        for name in varlist:
            if name in self.discover + ['total', 'lo']:
                ret.append(name)
        if not ret:
            raise Exception('No suitable network interfaces found to monitor')
        return ret

    def name(self):
        return ['pkt/'+name for name in self.vars]

    def extract(self):
        self.set2['total'] = [0, 0]
        for l in self.splitlines(replace=':'):
            if len(l) < 17: continue
            if l[2] == '0' and l[10] == '0': continue
            name = l[0]
            if name in self.vars :
                self.set2[name] = ( int(l[2]), int(l[10]) )
            if not self.totalfilter.match(name):
                self.set2['total'] = ( self.set2['total'][0] + int(l[2]), self.set2['total'][1] + int(l[10]))

        if update:
            for name in self.set2:
                self.val[name] = list(map(lambda x, y: (y - x) * 1.0 / elapsed, self.set1[name], self.set2[name]))

        if step == op.delay:
            self.set1.update(self.set2)



### Author: <gianfranco@mongodb.com>

global mongodb_user
mongodb_user = os.getenv('DOOL_MONGODB_USER') or os.getenv('USER')

global mongodb_pwd
mongodb_pwd = os.getenv('DOOL_MONGODB_PWD')

global mongodb_host
mongodb_host = os.getenv('DOOL_MONGODB_HOST') or '127.0.0.1:27017'

class dool_plugin(dool):
  """
  Plugin for MongoDB.
  """
  def __init__(self):
    global pymongo
    import pymongo

    try:
      self.m = pymongo.MongoClient(mongodb_host)
      if mongodb_pwd:
        self.m.admin.authenticate(mongodb_user, mongodb_pwd)
      self.db = self.m.admin
    except Exception as e:
      raise Exception('Cannot interface with MongoDB server: %s' % e)

    self.name    = 'mongodb con'
    self.nick    = ('curr', 'avail')
    self.vars    = ('connections.current', 'connections.available')
    self.type    = 'd'
    self.width   = 5
    self.scale   = 2
    self.lastVal = {}

  def extract(self):
    status = self.db.command("serverStatus")

    for name in self.vars:
      self.val[name] = (int(self.getDoc(status, name)))

  def getDoc(self, dic, doc):
    par = doc.split('.')
    sdic = dic
    for p in par:
      sdic = sdic.get(p)

    return sdic



### Author: Bert de Bruijn <bert+dstat$debruijn,be>

### VMware ESX kernel vmhba stats
### Displays kernel vmhba statistics on VMware ESX servers

# NOTE TO USERS: command-line plugin configuration is not yet possible, so I've
# "borrowed" the -D argument. 
# EXAMPLES:
# # dstat --vmkhba -D vmhba1,vmhba2,total
# # dstat --vmkhba -D vmhba0
# You can even combine the Linux and VMkernel diskstats (but the "total" argument
# will be used by both).
# # dstat --vmkhba -d -D sda,vmhba1

class dool_plugin(dool):
    def __init__(self):
        self.name = 'vmkhba'
        self.nick = ('read', 'writ')
        self.cols = 2

    def discover(self, *list):
    # discover will list all vmhba's found.
    # we might want to filter out the unused vmhba's (read stats, compare with ['0', ] * 13)
        ret = []
        try:
            list = os.listdir('/proc/vmware/scsi/')
        except:
            raise Exception('Needs VMware ESX')
        for name in list:
            for line in dopen('/proc/vmware/scsi/%s/stats' % name).readlines():
                l = line.split()
                if len(l) < 13: continue
                if l[0] == 'cmds': continue
                if l == ['0', ] * 13: continue
                ret.append(name)
        return ret

    def vars(self):
    # vars will take the argument list - when implemented - , use total, or will use discover + total
        ret = []
        if op.disklist:
            list = op.disklist
        #elif not op.full:
        #   list = ('total', )
        else:
            list = self.discover
            list.sort()
        for name in list:
            if name in self.discover + ['total']:
                ret.append(name)
        return ret

    def check(self): 
        try:
            os.listdir('/proc/vmware')
        except:
            raise Exception('Needs VMware ESX')
        info(1, 'The vmkhba module is an EXPERIMENTAL module.')

    def extract(self):
        self.set2['total'] = (0, 0)
        for name in self.vars:
            self.set2[name] = (0, 0)
        for name in os.listdir('/proc/vmware/scsi/'):
            for line in dopen('/proc/vmware/scsi/%s/stats' % name).readlines():
                l = line.split()
                if len(l) < 13: continue
                if l[0] == 'cmds': continue
                if l[2] == '0' and l[4] == '0': continue
                if l == ['0', ] * 13: continue
                self.set2['total'] = ( self.set2['total'][0] + int(l[2]), self.set2['total'][1] + int(l[4]) )
                if name in self.vars and name != 'total':
                    self.set2[name] = ( int(l[2]), int(l[4]) )

            for name in self.set2:
                self.val[name] = list(map(lambda x, y: (y - x) * 1024.0 / elapsed, self.set1[name], self.set2[name]))

        if step == op.delay:
            self.set1.update(self.set2)



### Author: Dag Wieers <dag@wieers.com>

class dool_plugin(dool):
    def __init__(self):
        self.name = 'rpc server'
        self.nick = ('call', 'erca', 'erau', 'ercl', 'xdrc')
        self.vars = ('calls', 'badcalls', 'badauth', 'badclnt', 'xdrcall')
        self.type = 'd'
        self.width = 5
        self.scale = 1000
        self.open('/proc/net/rpc/nfsd')

    def extract(self):
        for l in self.splitlines():
            if not l or l[0] != 'rpc': continue
            for i, name in enumerate(self.vars):
                self.set2[name] = int(l[i+1])

        for name in self.vars:
            self.val[name] = (self.set2[name] - self.set1[name]) * 1.0 / elapsed

        if step == op.delay:
            self.set1.update(self.set2)

# vim:ts=4:sw=4:et



### Author: Ming-Hung Chen <minghung.chen@gmail.com>


class dool_plugin(dool):
    bonddirname = '/proc/net/bonding/'
    netdirname = '/sys/class/net/'
    """
    Bytes received or sent through bonding interfaces.
    Usage:
        dool --bond -N <adapter name>,total
        default dool --bond is the same as
        dool --bond -N total

    """

    def __init__(self):
        self.nick = ('recv', 'send')
        self.type = 'd'
        self.cols = 2
        self.width = 6

    def discover(self, *objlist):
        ret = []
        for subdirname in os.listdir(self.bonddirname):
            if not os.path.isdir(os.path.join(self.netdirname,subdirname, 'statistics')) : continue
            device_dir =  os.path.join(self.netdirname, subdirname, 'statistics')
            ret.append(subdirname)
        ret.sort()
        for item in objlist: ret.append(item)
        return ret

    def vars(self):
        ret = []
        if op.netlist:
            varlist = op.netlist
        elif not op.full:
            varlist = ('total',)
        else:
            varlist = self.discover
            varlist.sort()
        for name in varlist:
            if name in self.discover + ['total']:
                ret.append(name)
        if not ret:
            raise Exception('No suitable network interfaces found to monitor')
        return ret

    def name(self):
        return ['bond/'+name for name in self.vars]

    def extract(self):
        self.set2['total'] = [0, 0]
        ifaces = self.discover
        for name in self.vars: self.set2[name] = [0, 0]
        for name in ifaces:
            rcv_counter_name=os.path.join(self.netdirname, name, 'statistics/rx_bytes')
            xmit_counter_name=os.path.join(self.netdirname, name, 'statistics/tx_bytes')
            rcv_lines = dopen(rcv_counter_name).readlines()
            xmit_lines = dopen(xmit_counter_name).readlines()
            if len(rcv_lines) < 1 or len(xmit_lines) < 1:
                continue
            rcv_value = int(rcv_lines[0])
            xmit_value = int(xmit_lines[0])
            if name in self.vars :
                self.set2[name] = (rcv_value, xmit_value)
            self.set2['total'] = ( self.set2['total'][0] + rcv_value, self.set2['total'][1] + xmit_value)
        if update:
            for name in self.set2:
                self.val[name] = [
                    (self.set2[name][0] - self.set1[name][0]) / elapsed,
                    (self.set2[name][1] - self.set1[name][1]) / elapsed,                    
                ]
                if self.val[name][0] < 0: self.val[name][0] += maxint + 1
                if self.val[name][1] < 0: self.val[name][1] += maxint + 1
        if step == op.delay:
            self.set1.update(self.set2)



### Author: Dag Wieers <dag$wieers,com>

class dool_plugin(dool):
    """
    The average size (in sectors) of the requests that were issued
    to the device.
    """

    def __init__(self):
        self.version = 2
        self.nick = ('avgrq',)
        self.type = 'f'
        self.width = 4
        self.scale = 10
        self.open('/proc/diskstats')
        self.cols = 1
        self.struct = dict( nr_ios=0, rd_sect=0, wr_sect=0 )

    def discover(self, *objlist):
        ret = []
        for l in self.splitlines():
            if len(l) < 13: continue
            if set(l[3:]) == {'0'}: continue
            name = l[2]
            ret.append(name)
        for item in objlist: ret.append(item)
        if not ret:
            raise Exception('No suitable block devices found to monitor')
        return ret

    def vars(self):
        ret = []
        if op.disklist:
            varlist = op.disklist
        else:
            varlist = []
            blockdevices = [os.path.basename(filename) for filename in glob.glob('/sys/block/*')]
            for name in self.discover:
                if DOOL_DISKFILTER.match(name): continue
                if name not in blockdevices: continue
                varlist.append(name)
            varlist.sort()
        for name in varlist:
            if name in self.discover:
                ret.append(name)
        return ret

    def name(self):
        return self.vars

    def extract(self):
        for l in self.splitlines():
            if len(l) < 13: continue
            if set(l[3:]) == {'0'}: continue
            if l[3] == '0' and l[7] == '0': continue
            name = l[2]
            if name not in self.vars or name == 'total': continue
            self.set2[name] = dict(
                nr_ios = int(l[3])+int(l[7]),
                rd_sect = int(l[9]),
                wr_sect = int(l[11]),
            )

        for name in self.vars:
            tput = ( self.set2[name]['nr_ios'] - self.set1[name]['nr_ios'] )
            if tput:
                ticks = self.set2[name]['rd_sect'] - self.set1[name]['rd_sect'] + \
                        self.set2[name]['wr_sect'] - self.set1[name]['wr_sect']
                self.val[name] = ( ticks * 1.0 / tput, )
            else:
                self.val[name] = ( 0.0, )

        if step == op.delay:
            self.set1.update(self.set2)



### Author: Dag Wieers <dag@wieers.com>

class dool_plugin(dool):
    def __init__(self):
        self.nick = ('fcnt', )
        self.type = 'd'
        self.width = 5
        self.scale = 1000
        self.open('/proc/user_beancounters')
        self.cols = 1 ### Is this correct ?

    def check(self):
        info(1, 'Module %s is still experimental.' % self.filename)

    def discover(self, *list):
        ret = []
        for l in self.splitlines():
            if len(l) < 7 or l[0] in ('uid', '0:'): continue
            ret.append(l[0][0:-1])
        ret.sort()
        for item in list: ret.append(item)
        return ret

    def name(self):
        ret = []
        for name in self.vars:
            if name == 'total':
                ret.append('total failcnt')
            else:
                ret.append(name)
        return ret

    def vars(self):
        ret = []
        if not op.full:
            list = ('total', )
        else: 
            list = self.discover
        for name in list: 
            if name in self.discover + ['total']:
                ret.append(name)
        return ret

    def extract(self):
        for name in self.vars + ['total']:
            self.set2[name] = 0
        for l in self.splitlines():
            if len(l) < 6 or l[0] == 'uid':
                continue
            elif len(l) == 7:
                name = l[0][0:-1]
                if name in self.vars:
                    self.set2[name] = self.set2[name] + int(l[6])
                self.set2['total'] = self.set2['total'] + int(l[6])
            elif name == '0':
                continue
            else:
                if name in self.vars:
                    self.set2[name] = self.set2[name] + int(l[5])
                self.set2['total'] = self.set2['total'] + int(l[5])

        for name in self.vars:
            self.val[name] = (self.set2[name] - self.set1[name]) * 1.0 / elapsed

        if step == op.delay:
            self.set1.update(self.set2)

# vim:ts=4:sw=4:et



### Authority: Dag Wieers <dag$wieers,com>

class dool_plugin(dool):
    """
    Top process with highest total latency.

    Displays name and total amount of CPU time waited in milliseconds of
    the process that has the highest total amount waited for the measured
    timeframe.

    For more information see:

        http://eaglet.rain.com/rick/linux/schedstat/
    """

    def __init__(self):
        self.name = 'highest total'
        self.vars = ('latency process',)
        self.type = 's'
        self.width = 17
        self.scale = 0
        self.pidset1 = {}

    def check(self):
        if not os.access('/proc/self/schedstat', os.R_OK):
            raise Exception('Kernel has no scheduler statistics [CONFIG_SCHEDSTATS], use at least 2.6.12')

    def extract(self):
        self.output = ''
        self.pidset2 = {}
        self.val['result'] = 0
        for pid in proc_pidlist():
            try:
                ### Reset values
                if pid not in self.pidset1:
                    self.pidset1[pid] = {'wait_ticks': 0}

                ### Extract name
                name = proc_splitline('/proc/%s/stat' % pid)[1][1:-1]

                ### Extract counters
                l = proc_splitline('/proc/%s/schedstat' % pid)
            except IOError:
                continue
            except IndexError:
                continue

            if len(l) != 3: continue

            self.pidset2[pid] = {'wait_ticks': int(l[1])}

            totwait = (self.pidset2[pid]['wait_ticks'] - self.pidset1[pid]['wait_ticks']) * 1.0 / elapsed

            ### Get the process that spends the most jiffies
            if totwait > self.val['result']:
                self.val['result'] = totwait
                self.val['pid'] = pid
                self.val['name'] = getnamebypid(pid, name)

        if step == op.delay:
            self.pidset1 = self.pidset2

        if self.val['result'] != 0.0:
            self.output = '%-*s%s' % (self.width-4, self.val['name'][0:self.width-4], cprint(self.val['result'], 'd', 4, 100))

        ### Debug (show PID)
#       self.output = '%*s %-*s' % (5, self.val['pid'], self.width-6, self.val['name'])

    def showcsv(self):
        return '%s / %.4f' % (self.val['name'], self.val['result'])

# vim:ts=4:sw=4:et



### Author: Dag Wieers <dag$wieers,com>

class dool_plugin(dool):
    def __init__(self):
        self.name = 'total cpu'
        self.vars = ( 'usr', 'sys', 'idl' )
        self.type = 'p'
        self.width = 3
        self.scale = 34
        self.server = os.getenv('DOOL_SNMPSERVER') or '192.168.1.1'
        self.community = os.getenv('DOOL_SNMPCOMMUNITY') or 'public'

    def check(self):
        try:
            global cmdgen
            from pysnmp.entity.rfc3413.oneliner import cmdgen
        except:
            raise Exception('Needs pysnmp and pyasn1 modules')

    def extract(self):
        self.set2['usr'] = int(snmpget(self.server, self.community, (1,3,6,1,4,1,2021,11,50,0)))
        self.set2['sys'] = int(snmpget(self.server, self.community, (1,3,6,1,4,1,2021,11,52,0)))
        self.set2['idl'] = int(snmpget(self.server, self.community, (1,3,6,1,4,1,2021,11,53,0)))
#        self.set2['usr'] = int(snmpget(self.server, self.community, (('UCD-SNMP-MIB', 'ssCpuRawUser'), 0)))
#        self.set2['sys'] = int(snmpget(self.server, self.community, (('UCD-SNMP-MIB', 'ssCpuRawSystem'), 0)))
#        self.set2['idl'] = int(snmpget(self.server, self.community, (('UCD-SNMP-MIB', 'ssCpuRawIdle'), 0)))

        if update:
            for name in self.vars:
                if sum(self.set2.values()) > sum(self.set1.values()):
                    self.val[name] = 100.0 * (self.set2[name] - self.set1[name]) / (sum(self.set2.values()) - sum(self.set1.values()))
                else:
                    self.val[name] = 0

        if step == op.delay:
            self.set1.update(self.set2)

# vim:ts=4:sw=4:et



### Author: Dag Wieers <dag@wieers.com>

#Version: 2.2
#VEID   user    nice    system   uptime     idle             strv   uptime          used           maxlat  totlat  numsched
#302    142926  0       10252    152896388  852779112954062  0      427034187248480 1048603937010  0       0       0
#301    27188   0       7896     152899846  853267000490282  0      427043845492614 701812592320   0       0       0

class dool_plugin(dool):
    def __init__(self):
        self.nick = ('usr', 'sys', 'idl', 'nic')
        self.type = 'p'
        self.width = 3
        self.scale = 34
        self.open('/proc/vz/vestat')
        self.cols = 4

    def check(self):
        info(1, 'Module %s is still experimental.' % self.filename)

    def discover(self, *list):
        ret = []
        for l in self.splitlines():
            if len(l) < 6 or l[0] == 'VEID': continue
            ret.append(l[0])
        ret.sort()
        for item in list: ret.append(item)
        return ret

    def name(self):
        ret = []
        for name in self.vars:
            if name == 'total':
                ret.append('total ve usage')
            else:
                ret.append('ve ' + name + ' usage')
        return ret

    def vars(self):
        ret = []
        if not op.full:
            list = ('total', )
        else: 
            list = self.discover
        for name in list: 
            if name in self.discover + ['total']:
                ret.append(name)
        return ret

    def extract(self):
        self.set2['total'] = [0, 0, 0, 0]
        for l in self.splitlines():
            if len(l) < 6 or l[0] == 'VEID': continue
            name = l[0]
            self.set2[name] = ( int(l[1]), int(l[3]), int(l[4]) - int(l[1]) - int(l[2]) - int(l[3]), int(l[2]) )
            self.set2['total'] = ( self.set2['total'][0] + int(l[1]), self.set2['total'][1] + int(l[3]), self.set2['total'][2] + int(l[4]) - int(l[1]) - int(l[2]) - int(l[3]), self.set2['total'][3] + int(l[2]) )

        for name in self.vars:
            for i in range(self.cols):
                self.val[name][i] = 100.0 * (self.set2[name][i] - self.set1[name][i]) / (sum(self.set2[name]) - sum(self.set1[name]))

        if step == op.delay:
            self.set1.update(self.set2)

# vim:ts=4:sw=4:et



### Author: Tom Van Looy <tom$ctors,net>

class dool_plugin(dool):
    """
    port of qmail_qstat to dstat
    """
    def __init__(self):
        self.name = 'qmail'
        self.nick = ('in_queue', 'not_prep')
        self.vars = ('mess', 'todo')
        self.type = 'd'
        self.width = 4
        self.scale = 100

    def check(self):
        for item in self.vars:
            if not os.access('/var/qmail/queue/'+item, os.R_OK):
                raise Exception('Cannot access qmail queues')

    def extract(self):
        for item in self.vars:
            self.val[item] = len(glob.glob('/var/qmail/queue/'+item+'/*/*'))

# vim:ts=4:sw=4:et



# Author: Roberto Polli <rpolli@redhat.com>
#
# NOTE: Edit the jcmd location according to your path or use update-alternatives.
global BIN_JCMD
BIN_JCMD = '/usr/bin/jcmd'


class dool_plugin(dool):
    """
    This plugin gathers jvm stats via jcmd.

    Usage:
       JVM_PID=15123 dstat --jvm-full 

    Minimize the impacts of jcmd and consider using:

        dstat --noupdate

    For full informations on jcmd see:

      - http://docs.oracle.com/javase/7/docs/technotes/tools/solaris/jcmd.html
      - https://docs.oracle.com/javase/8/docs/technotes/guides/troubleshoot/tooldescr006.html

    This requires the presence of /tmp/hsperfdata_* directory, so
     it WON'T WORK if you add -XX:-UsePerfData or -XX:+PerfDisableSharedMem.

    """

    def __init__(self):
        self.name = 'jvm_full'
        self.vars = ('clsL', 'clsU', 'fgc', 'heap', 'heap%',
                     'heapmax', 'perm', 'perm%', 'permmax')
        self.type = 'f'
        self.width = 5
        self.scale = 1000

    def check(self):
        """Preliminar checks. If no pid is passed, defaults to 0.
        """
        if not os.access(BIN_JCMD, os.X_OK):
            raise Exception('Needs jstat binary')

        try:
            self.jvm_pid = int(os.environ.get('JVM_PID',0))
        except Exception as e:
            self.jvm_pid = 0

        return True

    @staticmethod
    def _to_stat(k, v):
        try:
            return k, int(v)
        except (KeyError, ValueError, AttributeError):
            return k, v

    @staticmethod
    def _cmd_splitlines(cmd):
        """Splits a txt output of lines like key=value.
        """
        for l in os.popen(cmd):
            yield l.strip().split("=", 1)

    def extract(self):
        try:
            lines = self._cmd_splitlines(
                '%s %s PerfCounter.print ' % (BIN_JCMD, self.jvm_pid))
            table = dict(self._to_stat(*l) for l in lines
                         if len(l) > 1)
            if table:
                # Number of loaded classes.
                self.set2['clsL'] = table['java.cls.loadedClasses']
                self.set2['clsU'] = table['java.cls.unloadedClasses']
                # Number of Full Garbage Collection events.
                self.set2['fgc'] = table['sun.gc.collector.1.invocations']
                # The heap space is made up of Old Generation and Young
                # Generation (which is divided in Eden, Survivor0 and
                # Survivor1)
                self.set2['heap'] = table['sun.gc.generation.1.capacity'] + table[
                    'sun.gc.generation.0.capacity']
                # Usage is hidden in the nested spaces.
                self.set2['heapu'] = sum(table[k] for k in table
                                         if 'sun.gc.generation.' in k
                                         and 'used' in k)
                self.set2['heapmax'] = table['sun.gc.generation.1.maxCapacity'] + table[
                    'sun.gc.generation.0.maxCapacity']

                # Use PermGen on jdk7 and the new metaspace on jdk8
                try:
                    self.set2['perm'] = table['sun.gc.generation.2.capacity']
                    self.set2['permu'] = sum(table[k] for k in table
                                             if 'sun.gc.generation.2.' in k
                                             and 'used' in k)
                    self.set2['permmax'] = table[
                        'sun.gc.generation.2.maxCapacity']
                except KeyError:
                    self.set2['perm'] = table['sun.gc.metaspace.capacity']
                    self.set2['permu'] = table['sun.gc.metaspace.used']
                    self.set2['permmax'] = table[
                        'sun.gc.metaspace.maxCapacity']

            # Evaluate statistics on memory usage.
            for name in ('heap', 'perm'):
                self.set2[name + '%'] = 100 * self.set2[
                    name + 'u'] / self.set2[name]

            for name in self.vars:
                self.val[name] = self.set2[name]

            if step == op.delay:
                self.set1.update(self.set2)

        except IOError as e:
            if op.debug > 1:
                print('%s: lost pipe to jstat, %s' % (self.filename, e))
            for name in self.vars:
                self.val[name] = -1

        except Exception as e:
            if op.debug > 1:
                print('%s: exception' % e)
            for name in self.vars:
                self.val[name] = -1

# vim:ts=4:sw=4:et



### Author: David Nicklay <david-d$nicklay,com>
### Modified from disk-util: Dag Wieers <dag$wieers,com>

class dool_plugin(dool):
    """
    Read and Write average wait times of block devices.

    Displays the average read and write wait times of block devices
    """

    def __init__(self):
        self.nick = ('rawait', 'wawait')
        self.type = 'f'
        self.width = 4
        self.scale = 1
        self.open('/proc/diskstats')
        self.cols = 1
        self.struct = dict( rd_ios=0, wr_ios=0, rd_ticks=0, wr_ticks=0 )

    def discover(self, *objlist):
        ret = []
        for l in self.splitlines():
            if len(l) < 13: continue
            if set(l[3:]) == {'0'}: continue
            name = l[2]
            ret.append(name)
        for item in objlist: ret.append(item)
        if not ret:
            raise Exception('No suitable block devices found to monitor')
        return ret

    def vars(self):
        ret = []
        if op.disklist:
            varlist = op.disklist
        else:
            varlist = []
            blockdevices = [os.path.basename(filename) for filename in glob.glob('/sys/block/*')]
            for name in self.discover:
                if DOOL_DISKFILTER.match(name): continue
                if name not in blockdevices: continue
                varlist.append(name)
            varlist.sort()
        for name in varlist:
            if name in self.discover:
                ret.append(name)
        return ret

    def name(self):
        return self.vars

    def extract(self):
        for l in self.splitlines():
            if len(l) < 13: continue
            if l[5] == '0' and l[9] == '0': continue
            if set(l[3:]) == {'0'}: continue
            name = l[2]
            if name not in self.vars: continue
            self.set2[name] = dict(
                rd_ios = int(l[3]),
                wr_ios = int(l[7]),
                rd_ticks = int(l[6]),
                wr_ticks = int(l[10]),
            )

        for name in self.vars:
            rd_tput = self.set2[name]['rd_ios'] - self.set1[name]['rd_ios']
            wr_tput = self.set2[name]['wr_ios'] - self.set1[name]['wr_ios']
            if rd_tput:
                rd_wait = ( self.set2[name]['rd_ticks'] - self.set1[name]['rd_ticks'] ) * 1.0 / rd_tput
            else:
                rd_wait = 0
            if wr_tput:
                wr_wait = ( self.set2[name]['wr_ticks'] - self.set1[name]['wr_ticks'] ) * 1.0 / wr_tput
            else:
                wr_wait = 0
            self.val[name] = ( rd_wait, wr_wait )

        if step == op.delay:
            self.set1.update(self.set2)



### Author: Dag Wieers <dag@wieers.com>

class dool_plugin(dool):
    """
    Top most expensive I/O process

    Displays the name of the most expensive I/O process
    """
    def __init__(self):
        self.name = 'top-io'
        self.vars = ('    name    read  writ',)
        self.type = 's'
        self.width = 22
        self.scale = 0
        self.pidset1 = {}

    def check(self):
        if not os.access('/proc/self/io', os.R_OK):
            raise Exception('Kernel has no per-process I/O accounting [CONFIG_TASK_IO_ACCOUNTING], use at least 2.6.20')

    def extract(self):
        self.output       = ''
        self.pidset2      = {}
        self.val['usage'] = 0.0
        for pid in proc_pidlist():
            try:
                ### Reset values
                if pid not in self.pidset2:
                    self.pidset2[pid] = {'read_bytes:': 0, 'write_bytes:': 0}
                if pid not in self.pidset1:
                    self.pidset1[pid] = {'read_bytes:': 0, 'write_bytes:': 0}

                ### Extract name
                name = proc_splitline('/proc/%s/stat' % pid)[1][1:-1]

                ### Extract counters
                for l in proc_splitlines('/proc/%s/io' % pid):
                    if len(l) != 2: continue
                    self.pidset2[pid][l[0]] = int(l[1])
            except IOError:
                continue
            except IndexError:
                continue

            if (op.bits):
                factor = 8
            else:
                factor = 1

            read_usage  = (self.pidset2[pid]['read_bytes:']  - self.pidset1[pid]['read_bytes:'])  * factor / elapsed
            write_usage = (self.pidset2[pid]['write_bytes:'] - self.pidset1[pid]['write_bytes:']) * factor / elapsed
            usage       = read_usage + write_usage

            ### Get the process that spends the most jiffies
            if usage > self.val['usage']:
                self.val['usage']       = usage
                self.val['read_usage']  = read_usage
                self.val['write_usage'] = write_usage
                self.val['pid']         = pid
                self.val['name']        = getnamebypid(pid, name)

        if step == op.delay:
            self.pidset1 = self.pidset2

        if self.val['usage'] != 0.0:
            self.output = '%-*s%s %s' % (self.width-11, self.val['name'][0:self.width-11], cprint(self.val['read_usage'], 'd', 5, 1024), cprint(self.val['write_usage'], 'd', 5, 1024))

        ### Debug (show PID)
#        self.output = '%*s %-*s%s %s' % (5, self.val['pid'], self.width-17, self.val['name'][0:self.width-17], cprint(self.val['read_usage'], 'd', 5, 1024), cprint(self.val['write_usage'], 'd', 5, 1024))

    def showcsv(self):
        return '%s / %d:%d' % (self.val['name'], self.val['read_usage'], self.val['write_usage'])

# vim:ts=4:sw=4:et
